"""
PyTorch Geometric Algebra

An python library for Geometric Algebra in Pytorch
"""

__version__ = "0.0.2"
__author__ = 'Francesco Alesiani'
__credits__ = 'TF-GA (https://doi.org/10.5281/zenodo.3902404)'

from .torch_ga import GeometricAlgebra